#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import gobject
import gtk
import pango

import dialogs
import datatype
import mixedtree
import plist

class DataWidget(gtk.VBox):

  __gsignals__ = { "on-store"  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())}

  def __init__(self):
    gtk.VBox.__init__(self)
    
    frame = self.frame = gtk.Frame()
    
    label = gtk.Label()
    label.set_markup("<b>Data</b>")
    
    frame.set_label_widget(label)
    frame.set_shadow_type(gtk.SHADOW_NONE)

    self.pack_start(frame)
    self.buttons = None
    return

  def set_buttons(self, buttons):
    self.buttons = buttons
    buttons.connect("revert", self.revert)
    buttons.connect("store", self.store)
    buttons.show_all()

  def update(self, node):

    self.node = node

    if not self.is_node_editable():
      self.set_data_fixed()
    elif node.is_tensor(self.geometry_dim_tree):
      self.set_data_tensor()
    elif isinstance(node.datatype, tuple):
      self.set_data_combo()
    else:
      self.set_data_entry()

    return
 
  def revert(self, button = None):
    """
    "Revert Data" button click signal handler. Reverts data in the data frame.
    """

    self.update(self.node)

    return 
  
  def store(self, button = None):
    """
    "Store Data" button click signal handler. Stores data from the data frame
    in the treestore.
    """

    if not self.is_node_editable():
      return True
    elif self.node.is_tensor(self.geometry_dim_tree):
      return self.data_tensor_store()
    elif isinstance(self.node.datatype, tuple):
      return self.data_combo_store()
    else:
      return self.data_entry_store()

# This would be nice, look at it later
#    if self.scherror.errlist_is_open():
#      if self.scherror.errlist_type == 0:
#         self.scherror.on_validate_schematron()
#      else:
#         self.scherror.on_validate()

  def is_node_editable(self):
    return (self.node is not None
       and self.node.active
       and self.node.datatype is not None
       and self.node.datatype != "fixed"
       and (not self.node.is_tensor(self.geometry_dim_tree) or self.geometry_dim_tree.data is not None))
       # not A or B == A implies B
       # A B T
       # 0 0 1
       # 0 1 1
       # 1 0 0
       # 1 1 1

  def add_scrolled_window(self):
    scrolledWindow = gtk.ScrolledWindow()
    scrolledWindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    self.frame.add(scrolledWindow)
    scrolledWindow.show()
    return scrolledWindow

  def add_text_view(self):
    scrolledWindow = self.add_scrolled_window()
    self.set_child_packing(self.frame, True, True, 0, gtk.PACK_START)
    
    try:
      import gtksourceview2
      buf = gtksourceview2.Buffer()
      lang_manager = gtksourceview2.LanguageManager()
      buf.set_highlight_matching_brackets(True)
      if self.node is not None and self.node.is_code():
        codelanguage = self.node.get_code_language();
        if codelanguage in lang_manager.get_language_ids():
          language = lang_manager.get_language(codelanguage)
        else:
          language = lang_manager.get_language("python")
        buf.set_language(language)
        buf.set_highlight_syntax(True)
      textview = gtksourceview2.View(buffer=buf)
      textview.set_auto_indent(True)
      textview.set_insert_spaces_instead_of_tabs(True)
      textview.set_tab_width(2)
      if self.node is not None and self.node.is_code():
        textview.set_show_line_numbers(True)
        font_desc = pango.FontDescription("monospace")
        if font_desc:
          textview.modify_font(font_desc)
    except ImportError:
      textview = gtk.TextView()

    textview.set_pixels_above_lines(2)
    textview.set_pixels_below_lines(2)
    textview.set_wrap_mode(gtk.WRAP_WORD)
    textview.connect("focus-in-event", self.entry_focus_in)

    scrolledWindow.add(textview)
    textview.show()
    return textview

  def set_data_empty(self):
    """
    Empty the data frame.
    """

    if self.frame.child is not None:
      if isinstance(self.data, gtk.TextView):
        self.data.handler_block_by_func(self.entry_focus_in)
      elif isinstance(self.data, gtk.ComboBox):
        self.data.handler_block_by_func(self.combo_focus_child)

      self.frame.remove(self.frame.child)

    self.interacted = False

    return

  def set_data_fixed(self):
    """
    Create a non-editable text view to show help or fixed data.
    """
    
    self.set_data_empty()
   
    self.data = self.add_text_view()

    self.data.get_buffer().create_tag("tag")
    text_tag = self.data.get_buffer().get_tag_table().lookup("tag")

    self.data.set_cursor_visible(False)
    self.data.set_editable(False)
    self.buttons.hide()
    text_tag.set_property("foreground", "grey")

    if self.node is None:
      self.data.get_buffer().set_text("")
    elif not self.node.active:
      self.data.get_buffer().set_text("Inactive node")
    elif self.node.datatype is None:
      self.data.get_buffer().set_text("No data")
    elif self.node.is_tensor(self.geometry_dim_tree):
      self.data.get_buffer().set_text("Dimension not set")
    else: # self.node.datatype == "fixed":
      self.data.get_buffer().set_text(self.node.data)
 
    buffer_bounds = self.data.get_buffer().get_bounds()
    self.data.get_buffer().apply_tag(text_tag, buffer_bounds[0], buffer_bounds[1])

    return

  def set_data_entry(self):
    """
    Create a text view for data entry in the data frame.
    """

    self.set_data_empty()

    self.data = self.add_text_view()

    self.data.get_buffer().create_tag("tag")
    text_tag = self.data.get_buffer().get_tag_table().lookup("tag")
 
    self.data.set_cursor_visible(True)
    self.data.set_editable(True)
    self.buttons.show()

    if self.node.data is None:
      self.data.get_buffer().set_text(datatype.print_type(self.node.datatype))
      text_tag.set_property("foreground", "blue")
    else:
      self.data.get_buffer().set_text(self.node.data)

    buffer_bounds = self.data.get_buffer().get_bounds()
    self.data.get_buffer().apply_tag(text_tag, buffer_bounds[0], buffer_bounds[1])

    return

  def set_data_tensor(self):
    """
    Create a table container packed with appropriate widgets for tensor data entry
    in the node data frame.
    """

    self.set_data_empty()

    scrolledWindow = self.add_scrolled_window()

    dim1, dim2 = self.node.tensor_shape(self.geometry_dim_tree)
    self.data = gtk.Table(dim1, dim2)
    scrolledWindow.add_with_viewport(self.data)
    scrolledWindow.child.set_property("shadow-type", gtk.SHADOW_NONE)

    self.set_child_packing(self.frame, True, True, 0, gtk.PACK_START)

    self.show_all()
    self.buttons.show()

    is_symmetric = self.node.is_symmetric_tensor(self.geometry_dim_tree)
    for i in range(dim1):
      for j in range(dim2):
        iindex = dim1 - i - 1
        jindex = dim2 - j - 1

        entry = gtk.Entry()
        self.data.attach(entry, jindex, jindex + 1, iindex, iindex + 1)

        if not is_symmetric or i >= j:
          entry.show()
          entry.connect("focus-in-event", self.tensor_element_focus_in, jindex, iindex)

          if self.node.data is None:
            entry.set_text(datatype.print_type(self.node.datatype.datatype))
            entry.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("blue"))
          else:
            entry.set_text(self.node.data.split(" ")[jindex + iindex * dim2])

    self.interacted = [False for i in range(dim1 * dim2)]

    return

  def set_data_combo(self):
    """
    Create a combo box for node data selection in the node data frame. Add an
    entry if required.
    """

    self.set_data_empty()

    if isinstance(self.node.datatype[0], tuple):
      self.data = gtk.combo_box_entry_new_text()
    else:
      self.data = gtk.combo_box_new_text()

    self.frame.add(self.data)
    self.data.show()

    self.data.connect("set-focus-child", self.combo_focus_child)

    self.set_child_packing(self.frame, False, False, 0, gtk.PACK_START)

    if isinstance(self.node.datatype[0], tuple):
      self.buttons.show()
    else:
      self.buttons.hide()

    if self.node.data is None:
      if isinstance(self.node.datatype[0], tuple):
        self.data.child.set_text("Select " + datatype.print_type(self.node.datatype[1]) + "...")
      else:
        self.data.append_text("Select...")
        self.data.set_active(0)
      self.data.child.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("blue"))
      self.data.child.modify_text(gtk.STATE_PRELIGHT, gtk.gdk.color_parse("blue"))

    if isinstance(self.node.datatype[0], tuple):
      options = self.node.datatype[0]
    else:
      options = self.node.datatype

    for (i, opt) in enumerate(options):
      self.data.append_text(opt)
      if self.node.data == opt:
        self.data.set_active(i)

    if (isinstance(self.node.datatype[0], tuple)
       and self.node.data is not None
       and self.node.data not in self.node.datatype[0]):
      self.data.child.set_text(self.node.data)

    self.data.connect("changed", self.combo_changed)

    return

  def data_entry_store(self):
    """
    Attempt to store data read from a textview packed in the data frame.
    """

    new_data = self.data.get_buffer().get_text(*self.data.get_buffer().get_bounds())

    if new_data == "":
      return True

    if self.node.data is None and not self.interacted:
      return True
    else:
      value_check = self.node.validity_check(self.node.datatype, new_data)
      if value_check is None:
        dialogs.error(None, "Invalid value entered")
        return False
      elif value_check != self.node.data:
        self.node.set_data(value_check)
        if (isinstance(self.node, mixedtree.MixedTree)
           and "shape" in self.node.child.attrs.keys()
           and self.node.child.attrs["shape"][0] is int
           and isinstance(self.node.datatype, plist.List)
           and self.node.datatype.cardinality == "+"):
          self.node.child.set_attr("shape", str(len(value_check.split(" "))))

        self.emit("on-store")
        self.interacted = False

    return True

  def data_tensor_store(self):
    """
    Attempt to store data read from tensor data entry widgets packed in the 
    data frame.
    """

    dim1, dim2 = self.node.tensor_shape(self.geometry_dim_tree)
    is_symmetric = self.node.is_symmetric_tensor(self.geometry_dim_tree)

    if True not in self.interacted:
      return True

    entry_values = []
    for i in range(dim1):
      for j in range(dim2):
        if is_symmetric and i > j:
          entry_values.append(self.data.get_children()[i + j * dim1].get_text())
        else:
          entry_values.append(self.data.get_children()[j + i * dim2].get_text())

    changed = False
    for i in range(dim1):
      for j in range(dim2):
        if (self.interacted[j + i * dim2]
           and entry_values[j + i * dim2] != ""
           and (self.node.data is None
                or self.node.data.split(" ")[j + i * dim2] != entry_values[j + i * dim2])):
          changed = True
          break

    if not changed:
      return True
    elif (self.node.data is None and False in self.interacted) or "" in entry_values:
      dialogs.error(None, "Invalid value entered")
      return False

    new_data = ""
    for i in range(dim1):
      for j in range(dim2):
        new_data += " " + entry_values[j + i * dim2]

    value_check = self.node.validity_check(self.node.datatype, new_data)
    if value_check is None:
      return False
    elif not value_check == self.node.data:
      self.node.set_data(value_check)

      dim1, dim2 = self.node.tensor_shape(self.geometry_dim_tree)
      if int(self.node.child.attrs["rank"][1]) == 1:
        self.node.child.set_attr("shape", str(dim1))
      else:
        self.node.child.set_attr("shape", str(dim1) + " " + str(dim2))

      self.emit("on-store")
      self.interacted = [False for i in range(dim1 * dim2)]

    return True

  def data_combo_store(self):
    """
    Attempt to store data read from a combo box entry packed in the node data.
    """

    if not isinstance(self.node.datatype[0], tuple):
      return True

    new_data = self.data.get_text()

    if self.node.data is None and not self.interacted:
      return True
    elif not new_data in self.node.datatype[0]:
      new_data = self.node.validity_check(self.node.datatype[1], new_data)
      if new_data is None:
        return False

    if not new_data == self.node.data:
      self.node.set_data(new_data)
      self.emit("on-store")
      self.interacted = False

    return True
  
  def entry_focus_in(self, widget, event):
    """
    Called when a text view data entry widget gains focus. Used to delete the
    printable_type placeholder.
    """

    if (self.node is not None
       and self.node.datatype is not None
       and not self.node.is_tensor(self.geometry_dim_tree)
       and self.node.data is None
       and not self.interacted):
      self.data.get_buffer().set_text("")

    self.interacted = True

    return

  def tensor_element_focus_in(self, widget, event, row, col):
    """
    Called when a tensor data entry widget gains focus. Used to delete the
    printable_type placeholder.
    """

    dim1, dim2 = self.node.tensor_shape(self.geometry_dim_tree)
    if not self.interacted[row + col * dim2]:
      self.interacted[row + col * dim2] = True
      if self.node.is_symmetric_tensor(self.geometry_dim_tree):
        self.interacted[col + row * dim1] = True
      if self.node.data is None:
        widget.set_text("")
        widget.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("black"))

    return

  def combo_focus_child(self, container, widget):
    """
    Called when a data selection widget gains focus. Used to delete the select
    placeholder.
    """
    if not self.interacted:
      self.interacted = True
      if self.node.data is None:
        self.data.handler_block_by_func(self.combo_changed)
        if isinstance(self.node.datatype[0], tuple):
          self.data.child.set_text("")
        else:
          self.data.remove_text(0)

        self.data.child.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("black"))
        self.data.child.modify_text(gtk.STATE_PRELIGHT, gtk.gdk.color_parse("black"))
        self.data.handler_unblock_by_func(self.combo_changed)

    return

  def combo_changed(self, combo_box):
    """
    Called when a data combo box element is selected. Updates data in the
    treestore.
    """

    if not isinstance(self.node.datatype[0], tuple):
      text = self.data.get_active_text()
      if text is None:
        return

      self.node.set_data(text)
      self.emit("on-store")      
      self.interacted = False
    return

gobject.type_register(DataWidget)



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import gobject
import gtk

import attributewidget
import databuttonswidget
import datawidget
import mixedtree

class SliceView(gtk.Window):
  
  __gsignals__ = { "on-store" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
                   "update-name"  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())}

  def __init__(self, parent):
    gtk.Window.__init__(self)
    
    self.set_default_size(800, 600)
    self.set_title("Slice View")
    self.set_modal(True)
    self.set_transient_for(parent)

    mainvbox = gtk.VBox()
    self.vbox = gtk.VBox()

    scrolledWindow = gtk.ScrolledWindow()
    scrolledWindow.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
    scrolledWindow.add_with_viewport(self.vbox)

    self.databuttons = databuttonswidget.DataButtonsWidget()

    self.statusbar = gtk.Statusbar()

    mainvbox.pack_start(scrolledWindow, True)
    mainvbox.pack_start(self.databuttons, False)
    mainvbox.pack_start(self.statusbar, False)

    self.add(mainvbox)
    self.show_all()

  def update(self, node, tree):
    nodes = self.get_nodes(node, tree)
    if not nodes:
      self.destroy()

    for n in nodes:
      self.vbox.pack_start(self.control(n))

    maxwidth = 0
    for child in self.vbox.get_children():
      width, height = child.label.get_size_request()
      maxwidth = max(maxwidth, width)
    
    for child in self.vbox.get_children():
      child.label.set_size_request(maxwidth, -1)

    self.check_resize()

  def get_nodes(self, node, tree):
   nodes = []

   for child in tree.get_children():
     if child.active:
       if child.name == node.name and child.is_sliceable():
         nodes.append(child.get_mixed_data())
       nodes += self.get_nodes(node, child)
   
   return nodes

  def control(self, node):
    hbox = gtk.HBox()

    label = gtk.Label(node.get_name_path())
    hbox.label = label

    data = datawidget.DataWidget()
    data.geometry_dim_tree = self.geometry_dim_tree
    data.connect("on-store", self.on_store)
    data.set_buttons(self.databuttons)
    data.update(node)

    attributes = attributewidget.AttributeWidget()
    attributes.connect("on-store", self.on_store)
    attributes.connect("update-name", self.update_name)
    attributes.update(node)

    hbox.pack_start(label)
    hbox.pack_start(data)
    hbox.pack_start(attributes)
    
    hbox.show_all()

    return hbox

  def on_store(self, widget = None):
    self.emit("on-store")

  def update_name(self, widget = None):
    self.emit("update-name") 

gobject.type_register(SliceView)



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import gtk
import re
import TextBufferMarkup
import webbrowser

class DescriptionWidget(gtk.Frame):
  def __init__(self):
    gtk.Frame.__init__(self)

    scrolledWindow = gtk.ScrolledWindow()
    scrolledWindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)

    textView = self.textView = gtk.TextView()
    textView.set_editable(False)
    textView.set_wrap_mode(gtk.WRAP_WORD)
    textView.set_cursor_visible(False)

    textView.set_buffer(TextBufferMarkup.PangoBuffer())
    textView.connect("button-release-event", self.mouse_button_release)
    textView.connect("motion-notify-event", self.mouse_over)

    scrolledWindow.add(textView)

    label = gtk.Label()
    label.set_markup("<b>Description</b>")

    self.set_shadow_type(gtk.SHADOW_NONE)
    self.set_label_widget(label)
    self.add(scrolledWindow)

    return 

  def update(self, node):
    if node is None:
      self.set_description("<span foreground=\"grey\">No node selected</span>")
    elif node.doc is None:
      self.set_description("<span foreground=\"red\">No documentation</span>")
    else:
      self.set_description(node.doc)

    return

  def set_description(self, text):
    """
    Set the node description.
    """
    
    #text saved in self.text without markup
    self.text = text = self.render_whitespace(text)
    link_bounds = self.link_bounds = self.get_link_bounds(text)

    if link_bounds:
      new_text = []
      index = 0
      for bounds in link_bounds:  
        new_text.append(text[index:bounds[0]])
        new_text.append("<span foreground=\"blue\" underline=\"single\">")
        new_text.append(text[bounds[0]:bounds[1]])
        new_text.append("</span>")
        index = bounds[1]
             
      new_text.append(text[index:])
      text = ''.join(new_text)

    self.textView.get_buffer().set_text(text)

    return

  def get_link_bounds(self, text):
    """
    Return a list of tuples corresponding to the start and end points of links in
    the supplied string.
    """
    text = text.lower()
    bounds = []

    for match in re.finditer(r"\b(" #start at beginging of word
                                +r"(?:https?://|www\.)" #http:// https:// or www.
                                +r"(?:[a-z0-9][a-z0-9_\-]*[a-z0-9]\.)*" #Domains
                                +r"(?:[a-z][a-z0-9\-]*[a-z0-9])" #TLD
                                +r"(?:/([a-z0-9$_.+\\*'(),;:@&=\-]|%[0-9a-f]{2})*)*" #path
                                +r"(?:\?([a-z0-9$_.+!*'(),;:@&=\-]|%[0-9a-f]{2})*)?" #query
                                +r")", text):
      bounds.append(match.span())

    return bounds   

  def render_whitespace(self, desc):
    ''' Render the line wrapping in desc as follows:
    
    * Newlines followed by 0-1 spaces are ignored, and 1 space is used.
    * Blank lines start new paragraphs.
    * Newlines followed by more than 1 space are honoured.
    '''
    
    text = []
    para = False
    literal = False

    for line in desc.split("\n"):

      if line == "" or line.isspace(): #Blank line, start paragraph
        if literal: #if following a literal line add newlines
          text.append("\n")
        para = True

      elif line[0] == " " and line[1] == " ": # >1 space, treat literaly
        text.append("\n")
        text.append(line)
        para = False
        literal = True

      else: #normal case
        if para: #add if starting a new paragraph
          text.append("\n   ")
        if not line.startswith(" "):
          text.append(" ")
        text.append(line)
        para = False
        literal = False

    return ''.join(text)
 
  def get_hyperlink(self, x, y):
    """
    Given an x and y window position (eg from a mouse click) return the hyperlink
    at that position if there is one. Else return None.
    """
    if self.text is None:
      return None
    
    buffer_pos = self.textView.window_to_buffer_coords(gtk.TEXT_WINDOW_TEXT, x, y)
    char_offset = self.textView.get_iter_at_location(buffer_pos[0], buffer_pos[1]).get_offset()

    for bounds in self.link_bounds:
      if char_offset >= bounds[0] and char_offset <= bounds[1]:
        return self.text[bounds[0]:bounds[1]]

    return None
                         
  def mouse_over(self, widget, event):
    """
    Called when the mouse moves over the node description widget. Sets the cursor
    to a hand if the mouse hovers over a link.

    Based on code from HyperTextDemo class in hypertext.py from PyGTK 2.12 demos
    """

    if self.get_hyperlink(int(event.x), int(event.y)) is not None:
      self.textView.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND2))
    else:
      self.textView.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(gtk.gdk.Cursor(gtk.gdk.XTERM))

    return

  def mouse_button_release(self, widget, event):
    """
    Called when a mouse button is released over the node description widget.
    Launches a browser if the mouse release was over a link, the left mouse button
    was released and no text was selected.

    Based on code from HyperTextDemo class in hypertext.py from PyGTK 2.12 demos
    """

    if not event.button == 1:
      return

    if self.textView.get_buffer().get_selection_bounds():
      return

    hyperlink = self.get_hyperlink(int(event.x), int(event.y))
    if hyperlink is not None:
      webbrowser.open(hyperlink)

    return



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import gobject
import gtk

class DataButtonsWidget(gtk.HBox):

  __gsignals__ = { "revert" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
                   "store"  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())}

  def __init__(self):
    gtk.HBox.__gobject_init__(self)
    revertButton = gtk.Button()
    revertButton.set_label("Revert data")
    revertButton.connect("clicked", self._revert)

    storeButton = gtk.Button()
    storeButton.set_label("Store data")
    storeButton.connect("clicked", self._store)

    self.pack_start(revertButton)
    self.pack_end(storeButton)

    return

  def _revert(self, widget = None):
    self.emit("revert")

  def _store(self, widget = None):
    self.emit("store")

gobject.type_register(DataButtonsWidget)



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import gobject
import gtk

import datatype
import dialogs

class AttributeWidget(gtk.Frame):

  __gsignals__ = { "on-store" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
                   "update-name"  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())}

  def __init__(self):
    gtk.Frame.__init__(self)

    scrolledWindow = gtk.ScrolledWindow()
    scrolledWindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)

    treeview = self.treeview = gtk.TreeView()

    model = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)
    treeview.set_model(model)
    treeview.connect("motion-notify-event", self.treeview_mouse_over)

    key_renderer = gtk.CellRendererText()
    key_renderer.set_property("editable", False)

    column1 = gtk.TreeViewColumn("Name", key_renderer, text = 0)
    column1.set_cell_data_func(key_renderer, self.key_data_func)
    column1.set_property("min-width", 75)

    entry_renderer = gtk.CellRendererText()
    entry_renderer.connect("edited", self.entry_edited)
    entry_renderer.connect("editing-started", self.entry_edit_start)

    combo_renderer = gtk.CellRendererCombo()
    combo_renderer.set_property("text-column", 0)
    combo_renderer.connect("edited", self.combo_selected)
    combo_renderer.connect("editing-started", self.combo_edit_start)

    column2 = gtk.TreeViewColumn("Value", entry_renderer, text = 1)
    column2.pack_start(combo_renderer)
    column2.set_attributes(combo_renderer, text = 1)
    column2.set_cell_data_func(entry_renderer, self.entry_data_func)
    column2.set_cell_data_func(combo_renderer, self.combo_data_func)
    column2.set_property("expand", True)
    column2.set_property("min-width", 75)

    icon_renderer = gtk.CellRendererPixbuf()

    column3 = gtk.TreeViewColumn("", icon_renderer)
    column3.set_cell_data_func(icon_renderer, self.icon_data_func)

    treeview.append_column(column1)
    treeview.append_column(column2)
    treeview.append_column(column3)

    scrolledWindow.add(treeview)

    label = gtk.Label()
    label.set_markup("<b>Attributes</b>")

    self.set_label_widget(label)
    self.set_shadow_type(gtk.SHADOW_NONE)
    self.add(scrolledWindow)

  def update(self, node):

    self.treeview.get_model().clear()

    self.node = node

    if node is None or node.all_attrs_fixed():
      self.set_property("visible", False)
    else:
      self.set_property("visible", True)

      for key in node.attrs.keys():
        model = self.treeview.get_model()
        cell_model = gtk.ListStore(gobject.TYPE_STRING)

        iter = model.append()
        model.set_value(iter, 0, key)
        model.set_value(iter, 2, cell_model)

        if isinstance(node.attrs[key][0], tuple):
          if node.attrs[key][1] is None:
            if isinstance(node.attrs[key][0][0], tuple):
              model.set_value(iter, 1, "Select " + datatype.print_type(node.attrs[key][0][1]) + "...")
            else:
              model.set_value(iter, 1, "Select...")
          else:
            model.set_value(iter, 1, node.attrs[key][1])

          if isinstance(node.attrs[key][0][0], tuple):
            opts = node.attrs[key][0][0]
          else:
            opts = node.attrs[key][0]

          for opt in opts:
            cell_iter = cell_model.append()
            cell_model.set_value(cell_iter, 0, opt)

          self.treeview.get_column(2).set_property("visible", True)
        elif node.attrs[key][0] is None:
          model.set_value(iter, 1, "No data")
        elif node.attrs[key][1] is None:
          model.set_value(iter, 1, datatype.print_type(node.attrs[key][0]))
        else:
          model.set_value(iter, 1, node.attrs[key][1])

      self.treeview.queue_resize()

    return

  def treeview_mouse_over(self, widget, event):
    """
    Called when the mouse moves over the attributes widget. Sets the
    appropriate attribute widget tooltip.
    """

    path_info = self.treeview.get_path_at_pos(int(event.x), int(event.y))
    if path_info is None:
      try:
        self.treeview.set_tooltip_text("")
        self.treeview.set_property("has-tooltip", False)
      except:
        pass
      return

    path = path_info[0]
    col = path_info[1]
    if col is not self.treeview.get_column(1):
      try:
        self.treeview.set_tooltip_text("")
        self.treeview.set_property("has-tooltip", False)
      except:
        pass
      return

    iter = self.treeview.get_model().get_iter(path)
    iter_key = self.treeview.get_model().get_value(iter, 0)

    return

  def key_data_func(self, col, cell_renderer, model, iter):
    """
    Attribute name data function. Sets the cell renderer text colours.
    """

    iter_key = model.get_value(iter, 0)

    if not self.node.active or self.node.attrs[iter_key][0] is None or self.node.attrs[iter_key][0] == "fixed":
      cell_renderer.set_property("foreground", "grey")
    elif self.node.attrs[iter_key][1] is None:
      cell_renderer.set_property("foreground", "blue")
    else:
      cell_renderer.set_property("foreground", "black")

    return

  def entry_data_func(self, col, cell_renderer, model, iter):
    """
    Attribute text data function. Hides the renderer if a combo box is required,
    and sets colours and editability otherwise.
    """

    iter_key = model.get_value(iter, 0)

    if not self.node.active or self.node.attrs[iter_key][0] is None or self.node.attrs[iter_key][0] == "fixed":
      cell_renderer.set_property("editable", False)
      cell_renderer.set_property("foreground", "grey")
      cell_renderer.set_property("visible", True)
    elif not isinstance(self.node.attrs[iter_key][0], tuple):
      cell_renderer.set_property("editable", True)
      cell_renderer.set_property("visible", True)
      if self.node.attrs[iter_key][1] is None:
        cell_renderer.set_property("foreground", "blue")
      else:
        cell_renderer.set_property("foreground", "black")
    else:
      cell_renderer.set_property("editable", False)
      cell_renderer.set_property("visible", False)

    return

  def combo_data_func(self, col, cell_renderer, model, iter):
    """
    Attribute combo box data function. Hides the renderer if a combo box is not
    required, and sets the combo box options otherwise. Adds an entry if required.
    """

    iter_key = model.get_value(iter, 0)

    if self.node.active and isinstance(self.node.attrs[iter_key][0], tuple):
      cell_renderer.set_property("editable", True)
      cell_renderer.set_property("visible", True)
      if isinstance(self.node.attrs[iter_key][0][0], tuple):
        cell_renderer.set_property("has-entry", True)
      else:
        cell_renderer.set_property("has-entry", False)
      if self.node.attrs[iter_key][1] is None:
        cell_renderer.set_property("foreground", "blue")
      else:
        cell_renderer.set_property("foreground", "black")
    else:
      cell_renderer.set_property("visible", False)
      cell_renderer.set_property("editable", False)
    cell_renderer.set_property("model", model.get_value(iter, 2))

    return

  def icon_data_func(self, col, cell_renderer, model, iter):
    """
    Attribute icon data function. Used to add downward pointing arrows for combo
    attributes, for consistency with the LHS.
    """

    iter_key = model.get_value(iter, 0)

    if self.node.active and isinstance(self.node.attrs[iter_key][0], tuple):
      cell_renderer.set_property("stock-id", gtk.STOCK_GO_DOWN)
    else:
      cell_renderer.set_property("stock-id", None)

    return

  def entry_edit_start(self, cell_renderer, editable, path):
    """
    Called when editing is started on an attribute text cell. Used to delete the
    printable_type placeholder.
    """

    iter = self.treeview.get_model().get_iter(path)
    iter_key = self.treeview.get_model().get_value(iter, 0)

    if self.node.attrs[iter_key][1] is None:
      editable.set_text("")

    return

  def combo_edit_start(self, cell_renderer, editable, path):
    """
    Called when editing is started on an attribute combo cell. Used to delete the
    select placeholder for mixed entry / combo attributes.
    """

    iter = self.treeview.get_model().get_iter(path)
    iter_key = self.treeview.get_model().get_value(iter, 0)

    if isinstance(self.node.attrs[iter_key][0][0], tuple) and self.node.attrs[iter_key][1] is None:
      editable.child.set_text("")

    return

  def entry_edited(self, cell_renderer, path, new_text):
    """
    Called when editing is finished on an attribute text cell. Updates data in the
    treestore.
    """

    iter = self.treeview.get_model().get_iter(path)
    iter_key = self.treeview.get_model().get_value(iter, 0)

    if self.node.get_attr(iter_key) is None and new_text == "":
      return

    value_check = self.node.validity_check(self.node.attrs[iter_key][0], new_text)

    if value_check is not None and value_check != self.node.attrs[iter_key][1]:
      if iter_key == "name" and not self._name_check(value_check):
        return

      self.treeview.get_model().set_value(iter, 1, value_check)
      self.node.set_attr(iter_key, value_check)
      if iter_key == "name":
        self.emit("update-name")

      self.emit("on-store")

    return

  def combo_selected(self, cell_renderer, path, new_text):
    """
    Called when an attribute combo box element is selected, or combo box entry
    element entry is edited. Updates data in the treestore.
    """

    iter = self.treeview.get_model().get_iter(path)
    iter_key = self.treeview.get_model().get_value(iter, 0)

    if new_text is None:
      return

    if isinstance(self.node.attrs[iter_key][0][0], tuple) and new_text not in self.node.attrs[iter_key][0][0]:
      if self.node.get_attr(iter_key) is None and new_text == "":
        return

      new_text = self.node.validity_check(self.node.attrs[iter_key][0][1], new_text)
      if iter_key == "name" and not self._name_check(new_text):
        return False
    if new_text != self.node.attrs[iter_key][1]:
      self.treeview.get_model().set_value(iter, 1, new_text)
      self.node.set_attr(iter_key, new_text)
      if iter_key == "name":
        self.emit("update-name")

      self.emit("on-store")

    return

  def _name_check(self, value):
    """
    Check to see if the supplied data is a valid tree name.
    """

    valid_chars = "_:[]1234567890qwertyuioplkjhgfdsazxcvbnmMNBVCXZASDFGHJKLPOIUYTREWQ"
    for char in value:
      if char not in valid_chars:
        dialogs.error(None, "Invalid value entered")
        return False

    return True

gobject.type_register(AttributeWidget)



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import base64
import bz2
import copy
import StringIO
from lxml import etree


import gobject

import tree

class Choice(gobject.GObject):

  __gsignals__ = { "on-set-data" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (str,)),
                   "on-set-attr"  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (str, str))}

  def __init__(self, choices, schemaname="", cardinality=''):
    gobject.GObject.__init__(self)

    self.choices = choices
    if choices == []:
      raise Exception
    self.index = 0
    name = ""
    for choice in choices:
      assert choice.__class__ is tree.Tree
      choice.connect("on-set-data", self._on_set_data)
      choice.connect("on-set-attr", self._on_set_attr)

    name = ":".join(choice.name for choice in choices)

    self.name = name
    self.schemaname = schemaname
    self.cardinality = cardinality
    self.parent = None
    self.set_default_active()

  def _on_set_data(self, node, data):
    self.emit("on-set-data", data)

  def _on_set_attr(self, node, attr, value):
    self.emit("on-set-attr", attr, value)

  def get_attrs(self):
    """Get all attributes"""
    return self.get_current_tree().get_attrs()

  def set_default_active(self):
    self.active = True
    if self.cardinality == '?' or self.cardinality == '*':
      self.active = False

  def choose(self, i):
    self.index = i

  def find_tree(self, name):
    for t in self.choices:
      if t.name == name:
        return t

    debug.deprint("self.name == %s" % self.name, 0)
    for choice in self.choices:
      debug.deprint("choice.name == %s" % choice.name, 0)
    raise Exception, "No such choice name: %s" % name

  def set_active_choice_by_name(self, name):
    matched = False
    for t in self.choices:
      if t.name == name.strip():
        matched = True
        self.index = self.choices.index(t)

    if not matched:
      raise Exception, "no such name %s found" % name

    self.recompute_validity()

  def set_active_choice_by_ref(self, ref):
    self.index = self.choices.index(ref)
    self.recompute_validity()

  def get_current_tree(self):
    return self.choices[self.index]

  def add_children(self, schema):
    return self.get_current_tree().add_children(schema)

  def pickle(self):
    return base64.b64encode(bz2.compress(pickle.dumps(self)))

  def recompute_validity(self):
    self.get_current_tree().recompute_validity()

  def copy(self):
    new_choices = []
    for choice in self.choices:
      new_choices.append(choice.copy())

    new_choice = Choice(new_choices)
    for attr in ["index", "name", "schemaname", "cardinality", "active"]:
      setattr(new_choice, attr, copy.copy(getattr(self, attr)))

    new_choice.set_parent(self.parent)
    for choice in new_choice.choices:
      choice.children = copy.copy([])

    return new_choice

  def get_possible_names(self):
    return [x.name for x in self.choices]

  def set_parent(self, parent):
    self.parent = parent
    for choice in self.choices:
      choice.parent = parent

  def write_core(self, parent):
    self.choices[self.index].write_core(parent)
    return parent

  def is_comment(self):
    return False

  def get_comment(self):
    return None

  def get_display_name(self):
    """
    This is a fluidity hack, allowing the name displayed in the treeview on the
    left to be different to the element name. If it has an attribute name="xxx",
    element_tag (xxx) is displayed.
    """

    return self.get_current_tree().get_display_name()

  def get_name(self):
    return self.get_current_tree().get_name()

  def get_children(self):
    return [self.get_current_tree()]

  def get_choices(self):
    return self.choices

  def is_hidden(self):
    """
    Tests whether the supplied choice should be hidden in view.
    """
    return False

  def get_name_path(self, leaf = True):
    return self.get_current_tree().get_name_path(leaf)

  def get_mixed_data(self):
    return self

  def is_sliceable(self):
    return self.get_current_tree().is_sliceable()
 
  def __str__(self):
    """
    Returns the display name of the selected tree.
    """
    return self.get_display_name()

  def __repr__(self):
    return self.get_name_path() + "[" + self.name + "]"
 
gobject.type_register(Choice)



#!/usr/bin/env python                                                                                                                                                                                          

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

from lxml import etree
import os
import os.path
import debug
import sys
import copy
import urllib2

def preprocess(schemafile):
  p = etree.XMLParser(remove_comments=True)
  ns = 'http://relaxng.org/ns/structure/1.0'

  if 'http' in schemafile:
    schemafile_handle = urllib2.urlopen(schemafile)
  else:
    schemafile_handle = open(schemafile)
  
  try:
  	tree = etree.parse(schemafile_handle, p)
  except Exception:
  	debug.deprint("Error: %s is not a valid Relax NG schema" % schemafile, 0)
  	sys.exit(1)

  #
  # deal with include
  #
  includes = tree.xpath('/t:grammar//t:include', namespaces={'t': ns})

  for include in includes:
    include_parent = include.getparent()
    include_index = list(include_parent).index(include)

    # find the file
    file = None
    filename = include.attrib["href"]
    possible_files = [os.path.join(os.path.dirname(schemafile), filename), filename]
    possible_files.append(os.path.join("/usr/local/share/spud", filename))
    possible_files.append(os.path.join(os.path.dirname(__file__) + "/../../schema", filename))

    for possible_file in possible_files:
      try:
        if 'http' in possible_file:
          file = urllib2.urlopen(possible_file)
        else:
          file = open(possible_file)
        break
      except IOError:
        debug.deprint("IOError when searching for included file " + filename, 1)

    if file is None:
      debug.deprint("Error: could not locate included file %s" % filename, 0)
      debug.deprint("Path: %s" % possible_files)
      sys.exit(1)

    # parse the included xml file and steal all the nodes
    include_tree = etree.parse(file, p)
    nodes_to_take = include_tree.xpath('/t:grammar/*', namespaces={'t': ns})

    # here's where the magic happens:
    for node in nodes_to_take:
      include_parent.insert(include_index, copy.deepcopy(node))

    # now delete the include:
    include_parent.remove(include)

  grammar_list = tree.xpath('/t:grammar', namespaces={'t': ns})
  
  # If the .rnc didn't include a start = prefix, then no valid
  # grammar tag will be present. Let the user know.
  
  if len(grammar_list) == 0:
  		debug.deprint("Error: No grammar tag present in schema.", 0)
	  	sys.exit(1)

  grammar = grammar_list[0]

  defines = {}
  define_nodes = tree.xpath('/t:grammar//t:define', namespaces={'t': ns})

  #
  # deal with combine="interleave"
  #

  # first, fetch all the plain definitions
  for define in define_nodes:
    if "combine" not in define.attrib:
      name = define.attrib["name"]
      defines[name] = define

  # now look for interleaves with those
  for define in define_nodes:
    if "combine" in define.attrib and define.attrib["combine"] == "interleave":
      name = define.attrib["name"]
      if name not in defines:
        defines[name] = define
      else:
        matching_defn = defines[name]
        for child in define:
          matching_defn.append(copy.deepcopy(child))
  
  #
  # deal with combine="choice"
  #
  combine_names = []
  for define in define_nodes:
    if "combine" in define.attrib and define.attrib["combine"] == "choice":
      name = define.attrib["name"]
      combine_names.append(name)

  combine_names = list(set(combine_names))
  for name in combine_names:
    xpath = tree.xpath('/t:grammar//t:define[@name="%s"]' % name, namespaces={'t': ns})
    choices = []
    for node in xpath:
      choices = choices + list(node)
    define = etree.Element("define")
    define.attrib["name"] = name
    choice = etree.Element("choice")
    define.append(choice)
    for x in choices:
      choice.append(x)
    defines[name] = define

  # delete all the define nodes from the xml
  for define in define_nodes:
    parent = define.getparent()
    parent.remove(define)
  
  # add the modified defines back to the grammar
  for define in defines.values():
    grammar.append(define)

  return etree.tostring(tree, xml_declaration=True, encoding='utf-8', pretty_print=True)

if __name__ == "__main__":
  import sys
  schemafile = sys.argv[1]
  newfile = schemafile.replace(".rng", ".pp.rng")
  f = open(newfile, "w")
  f.write(preprocess(schemafile))



import os
import os.path
import sys
import thread
import traceback

import gtk.gdk

import debug

plugins = []

class PluginDetails(object):
  def __init__(self, applies, name, cb):
    self.applies = applies
    self.name = name
    self.cb = cb

  def matches(self, xpath):
    try:
      return self.applies(xpath)
    except Exception:
      debug.deprint("Warning: plugin %s raised an exception in matching function." % self.name, 0)
      return False

  def execute(self, xml, xpath):
    thread.start_new_thread(self.cb, (xml, xpath))

def register_plugin(applies, name, cb):
  global plugins
  p = PluginDetails(applies, name, cb)
  plugins.append(p)

def configure_plugins(suffix):
  homedir = os.path.expanduser('~')
  dirs=[]
  if "DIAMOND_CONFIG_PATH" in os.environ:
    dirs += os.environ["DIAMOND_CONFIG_PATH"].split(":")
  dirs.append(os.path.join(homedir, ".diamond", "plugins", suffix))
  dirs.append("/usr/local/share/diamond/plugins/" + suffix)
  if sys.platform != "win32" and sys.platform != "win64":
    dirs.append("/etc/diamond/plugins/" + suffix)

  for dir in dirs:
    sys.path.insert(0, dir)
    try:
      for file in os.listdir(dir):
        module_name, ext = os.path.splitext(file)
        if ext == ".py":
          try:
            debug.deprint("Attempting to import " + module_name, 1)
            module = __import__(module_name)
          except:
            debug.deprint("Plugin raised an exception:", 0)
            tb = traceback.format_exception(sys.exc_info()[0] ,sys.exc_info()[1], sys.exc_info()[2])
            tb_msg = ""
            for tbline in tb:
              tb_msg += tbline
            debug.deprint(tb_msg, 0)
    except OSError:
      pass

def cb_decorator(f):
  def wrapper(*args, **kwargs):
    gtk.gdk.threads_enter()

    try:
      f(*args, **kwargs)
    except:
      debug.deprint("Plugin raised an exception:", 0)
      tb = traceback.format_exception(sys.exc_info()[0] ,sys.exc_info()[1], sys.exc_info()[2])
      tb_msg = ""
      for tbline in tb:
        tb_msg += tbline
      debug.deprint(tb_msg, 0)

    gtk.gdk.threads_leave()
  return wrapper



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import os
import os.path
import sys
import cStringIO as StringIO

import gobject
import gtk
import threading

from lxml import etree

import attributewidget
import databuttonswidget
import datawidget
import mixedtree

from config import config
import dxdiff.diff as xmldiff

class DiffView(gtk.Window):

  def __init__(self, path, tree):
    gtk.Window.__init__(self)

    self.__add_controls()

    tree1 = etree.parse(path)
    tree2 = etree.ElementTree(tree.write_core(None))

    self.__update(tree1, tree2)

    self.show_all()

  def __add_controls(self):
    self.set_default_size(800, 600)
    self.set_title("Diff View")

    mainvbox = gtk.VBox()

    menubar = gtk.MenuBar()
    edititem = gtk.MenuItem("_Edit")
    menubar.append(edititem)

    agr = gtk.AccelGroup()
    self.add_accel_group(agr)

    self.popup = editmenu = gtk.Menu()
    edititem.set_submenu(editmenu)
    copyitem = gtk.MenuItem("Copy")
    copyitem.connect("activate", self.on_copy)
    key, mod = gtk.accelerator_parse("<Control>C")
    copyitem.add_accelerator("activate", agr, key, mod, gtk.ACCEL_VISIBLE)
    editmenu.append(copyitem)

    mainvbox.pack_start(menubar, expand = False)

    hpane = gtk.HPaned()
    scrolledwindow = gtk.ScrolledWindow()
    scrolledwindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    
    self.treeview = gtk.TreeView()

    self.treeview.get_selection().set_mode(gtk.SELECTION_SINGLE)
    self.treeview.get_selection().connect("changed", self.on_select_row)

    # Node column
    celltext = gtk.CellRendererText()
    column = gtk.TreeViewColumn("Node", celltext)
    column.set_cell_data_func(celltext, self.set_celltext)

    self.treeview.append_column(column)

    # 0: The node tag
    # 1: The attributes dict
    # 2: The value of the node if any
    # 3: The old value of the node
    # 4: "insert", "delete", "update",  ""
    self.treestore = gtk.TreeStore(gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)
    self.treeview.set_enable_search(False)
    self.treeview.connect("button_press_event", self.on_treeview_button_press)
    self.treeview.connect("popup_menu", self.on_treeview_popup)
    scrolledwindow.add(self.treeview)
    hpane.pack1(scrolledwindow, True)

    vpane = gtk.VPaned()
    frame = gtk.Frame()
    label = gtk.Label()
    label.set_markup("<b>Attributes</b>")
    frame.set_label_widget(label)
    frame.set_shadow_type(gtk.SHADOW_NONE)

    self.attribview = gtk.TreeView()

    celltext = gtk.CellRendererText()
    keycolumn = gtk.TreeViewColumn("Key", celltext)
    keycolumn.set_cell_data_func(celltext, self.set_cellkey)

    self.attribview.append_column(keycolumn)

    celltext = gtk.CellRendererText()
    valuecolumn = gtk.TreeViewColumn("Value", celltext)
    valuecolumn.set_cell_data_func(celltext, self.set_cellvalue)

    self.attribview.append_column(valuecolumn)

    frame.add(self.attribview)
    vpane.pack1(frame)

    frame = gtk.Frame()
    label = gtk.Label()
    label.set_markup("<b>Data</b>")
    frame.set_label_widget(label)
    frame.set_shadow_type(gtk.SHADOW_NONE)

    self.dataview = gtk.TextView()
    self.dataview.set_cursor_visible(False)
    self.dataview.set_editable(False)
    self.__create_tags(self.dataview.get_buffer())

    frame.add(self.dataview)
    vpane.pack2(frame)

    hpane.pack2(vpane)
    mainvbox.pack_start(hpane)
    self.add(mainvbox)

  def on_treeview_button_press(self, treeview, event):
    pathinfo = treeview.get_path_at_pos(int(event.x), int(event.y))
    if event.button == 3:
      if pathinfo is not None:
        treeview.get_selection().select_path(pathinfo[0])
        self.show_popup(None, event.button, event.time)
        return True

  def popup_location(self, widget, user_data):
    column = self.treeview.get_column(0)
    path = self.treeview.get_selection().get_selected()[1]
    area = self.treeview.get_cell_area(path, column)
    tx, ty = area.x, area.y
    x, y = self.treeview.tree_to_widget_coords(tx, ty)
    return (x, y, True)

  def on_treeview_popup(self, treeview):
    self.show_popup(None, self.popup_location, gtk.get_current_event_time())
    return

  def show_popup(self, func, button, time):
    self.popup.popup( None, None, func, button, time)
    return

  def __update(self, tree1, tree2):

    def async_diff(self, tree1, tree2):
      editscript = xmldiff.diff(tree1, tree2)
      self.__set_treestore(tree1.getroot())
      self.__parse_editscript(editscript)
      self.__floodfill(self.treestore.get_iter_root())
      gtk.idle_add(self.treeview.set_model, self.treestore) 

    t = threading.Thread(target = async_diff, args = (self, tree1, tree2))
    t.start()

  def __set_treestore(self, tree, iter = None):

    attrib = {}
    for key, value in tree.attrib.iteritems():
      #             (new,   old,   edit)
      attrib[key] = (value, value, None)

    child_iter = self.treestore.append(iter, [tree.tag, attrib, tree.text, tree.text, None]) 
    for child in tree:
      self.__set_treestore(child, child_iter)

  def __parse_editscript(self, editscript):
    for edit in editscript:
      iter, key = self.__get_iter(edit["location"])
      if key:
        attribs = self.treestore.get_value(iter, 1)
        old = attribs[key][1]
        if edit["type"] == "delete":
          attribs[key] = (None, old, "delete")
        elif edit["type"] == "update":
          attribs[key] = (edit["value"], old, "update")
        elif edit["type"] == "move":
          attribs[key] = (None, old, "delete")
          self.__insert(self.__get_iter(edit["value"])[0], key + " " + old, 0)

      else:

        if edit["type"] == "insert":
          self.__insert(iter, edit["value"], int(edit["index"]))
        elif edit["type"] == "delete":
          self.treestore.set(iter, 2, None)
          self.treestore.set(iter, 4, "delete")
        elif edit["type"] == "update":
          self.treestore.set(iter, 2, edit["value"])
        elif edit["type"] == "move":
          self.__move(iter, edit["value"], int(edit["index"]))

  def __floodfill(self, iter, parentedit = None):
    """
    Floodfill the tree with the correct edit types.
    If something has changed below you, "subupdate"
    If your value or attrs has changed "update"
    If insert, all below insert
    If delete, all below delete
    """
    attribs, new, old, edit = self.treestore.get(iter, 1, 2, 3, 4)

    if parentedit == "insert":
      edit = "insert"
    elif parentedit == "delete":
      edit = "delete"

    if edit == "insert" or edit == "delete":
      for key, (valuenew, valueold, valueedit) in attribs.iteritems():
        attribs[key] = (valuenew, valueold, edit)

      child = self.treestore.iter_children(iter)
      while child is not None:
        self.__floodfill(child, edit)
        child = self.treestore.iter_next(child)

      self.treestore.set(iter, 4, edit)
    else:
      update = False
      for key in attribs:
        # edit value
        if attribs[key][2] is not None:
          update = True
          break
      if new != old:
        update = True

      if update:
        self.treestore.set(iter, 4, "update")
      else:
        child = self.treestore.iter_children(iter)
        while child is not None:
          change = self.__floodfill(child, edit)
          if change is not None:
            self.treestore.set(iter, 4, "subupdate")
          child = self.treestore.iter_next(child)

    return self.treestore.get_value(iter, 4)

  def __insert(self, iter, value, index):
    if " " in value:
      key, value = value.split(" ")
      attrib = self.treestore.get_value(iter, 1)
      attrib[key] = (value, None, "insert")
    else:
      before = self.__iter_nth_child(iter, index - 1)
      if before:
        self.treestore.insert_before(iter, before, [value, {}, None, None, "insert"]) 
      else:
        self.treestore.append(iter, [value, {}, None, None, "insert"])

  def __move(self, iter, value, index):
    """
    Copy the entire subtree at iter to the path at value[index],
    mark all of iter as deleted, and all of the copy inserted.
    """
    tag, attrib, text = self.treestore.get(iter, 0, 1, 2)
    self.treestore.set(iter, 2, None)
    self.treestore.set(iter, 4, "delete")

    destiter = self.__get_iter(value)[0]

    before = self.__iter_nth_child(destiter, index - 1)
    if before:
      destiter = self.treestore.insert_before(destiter, before, [tag, attrib, text, None, "insert"]) 
    else:
      destiter = self.treestore.append(destiter, [tag, attrib, text, None, "insert"])

    def move(iterfrom, iterto):
      for childfrom in self.__iter_children(iterfrom):
        tag, attrib, text = self.treestore.get(childfrom, 0, 1, 2)
        self.treestore.set(childfrom, 2, None)
        self.treestore.set(childfrom, 4, "delete")

        childto = self.treestore.append(iterto, [tag, attrib, text, None, "insert"])
        move(childfrom, childto)

    move(iter, destiter)

  def __iter_children(self, iter):
    child = self.treestore.iter_children(iter)

    while child:
      active = self.treestore.get_value(child, 4) != "delete"
      if active:
        yield child
      child = self.treestore.iter_next(child)

  def __iter_nth_child(self, iter, n):

    for child in self.__iter_children(iter):
      if n == 0:
        return child
      else:
        n -= 1
    return None

  def __get_iter(self, path, iter = None):
    """
    Convert the given XML path to an iter into the treestore.
    """

    if iter is None:
      iter = self.treestore.get_iter_first()

    tag, edit = self.treestore.get(iter, 0, 4)
    if edit == "delete":
      return None # don't search deleted paths

    parentiter = self.treestore.iter_parent(iter)
    if parentiter:
      siblings = []
      for siblingiter in self.__iter_children(parentiter):
        siblingtag = self.treestore.get_value(siblingiter, 0)
        if siblingtag == tag:
          siblings.append(self.treestore.get_path(siblingiter))

      if len(siblings) != 1:
        index = "[" + str(siblings.index(self.treestore.get_path(iter)) + 1) + "]"
      else:
        index = ""

      tag = "/" + tag + index
    else:
      tag = "/" + tag

    index = path.find("/", 1)
    if index == -1:
      index = len(path)

    root = path[:index]
    path = path[index:]

    #check we match root
    if root != tag:
      return None

    if path:
      # check for text()
      if path == "/text()":
        return (iter, None)

      # check attributes
      if path.startswith("/@"):
        attrib = self.treestore.get_value(iter, 1)
        for key in attrib:
          if path == "/@" + key:
            return (iter, key)
        return None

      # check children
      for iter in self.__iter_children(iter):
        edit = self.treestore.get_value(iter, 4)
        if edit != "delete":
          result = self.__get_iter(path, iter)
          if result:
            return result

      return None
    else:
      # must be us
      return (iter, None)

  def on_select_row(self, selection):
    """
    Called when a row is selected.
    """
    (model, row) = selection.get_selected()
    if row is None:
      return

    attrib, new, old, edit = model.get(row, 1, 2, 3, 4)

    databuffer = self.dataview.get_buffer()
    tag = databuffer.get_tag_table().lookup("tag")
    if new or old: 
      self.__set_textdiff(self.dataview.get_buffer(), old, new)
      tag.set_property("background-set", False)
      tag.set_property("foreground", "black")
    else:
      databuffer.set_text("No data")
      self.__set_cell_property(tag, None)
      tag.set_property("foreground", "grey")

    bounds = databuffer.get_bounds()
    databuffer.apply_tag(tag, bounds[0], bounds[1])

    # 0: Key
    # 1: Value
    # 2: Old value
    # 3: "insert", "delete", "update",  ""

    attribstore = gtk.TreeStore(gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)

    for key, (new, old, diff) in attrib.iteritems():
      attribstore.append(None, [key, new, old, diff])

    self.attribview.set_model(attribstore)

  def __set_textdiff(self, databuffer, old, new):
    text1 = old.splitlines() if old else []
    text2 = new.splitlines() if new else []

    from difflib import Differ
    differ = Differ()
    result = differ.compare(text1, text2)
    result = [line + "\n" for line in result if not line.startswith("? ")]

    databuffer.set_text("") 
    for line in result:
      iter = databuffer.get_end_iter()
      if line.startswith("  "):
        databuffer.insert(iter, line)
      elif line.startswith("+ "):
        databuffer.insert_with_tags_by_name(iter, line, "add")
      elif line.startswith("- "):
        databuffer.insert_with_tags_by_name(iter, line, "rem") 

   
  def __create_tags(self, databuffer):
    databuffer.create_tag("tag")
    add = databuffer.create_tag("add")
    rem = databuffer.create_tag("rem")

    add.set_property("background", config.get("colour", "diffadd"))
    rem.set_property("background", config.get("colour", "diffsub"))

  def __set_cell_property(self, cell, edit):
    if edit is None:
      cell.set_property("foreground", config.get("colour", "normal"))
    elif edit == "insert":
      cell.set_property("foreground", config.get("colour", "insert"))
    elif edit == "delete":
      cell.set_property("foreground", config.get("colour", "delete"))
    elif edit == "update":
      cell.set_property("foreground", config.get("colour", "update"))
    elif edit == "subupdate":
      cell.set_property("foreground", config.get("colour", "subupdate"))

  def set_celltext(self, column, cell, model, iter):
  
    tag, text, edit = model.get(iter, 0, 2, 4)

    cell.set_property("text", tag)
    self.__set_cell_property(cell, edit) 

  def set_cellkey(self, column, cell, model, iter):
    
    key, edit = model.get(iter, 0, 3)
    cell.set_property("text", key)
    self.__set_cell_property(cell, edit) 

  def set_cellvalue(self, column, cell, model, iter):
 
    new, old, edit = model.get(iter, 1, 2, 3)
    if edit == "delete":
      cell.set_property("text", old)
    else:
      cell.set_property("text", new)
    self.__set_cell_property(cell, edit) 

  def _get_focus_widget(self, parent):
    """
    Gets the widget that is a child of parent with the focus.
    """
    focus = parent.get_focus_child()
    if focus is None or (focus.flags() & gtk.HAS_FOCUS):
      return focus
    else:
      return self._get_focus_widget(focus)

  def _handle_clipboard(self, widget, signal):
    """
    This finds the currently focused widget.
    If no widget is focused or the focused widget doesn't support 
    the given clipboard operation use the treeview (False), otherwise
    signal the widget to handel the clipboard operation (True).
    """
    widget = self._get_focus_widget(self)

    if widget is None or widget is self.treeview:
      return False

    if gobject.signal_lookup(signal + "-clipboard", widget):
      widget.emit(signal + "-clipboard")
      return True
    else:
      return False

  def __get_treestore(self, iter):

    tag, attrib, text = self.treestore.get(iter, 0, 1, 2)

    tree = etree.Element(tag)

    for key, (newvalue, oldvalue, edit) in attrib.iteritems():
      tree.attrib[key] = newvalue

    child_iter = self.treestore.iter_children(iter)
    while child_iter:
      child = self.__get_treestore(child_iter)
      tree.append(child)
      child_iter = self.treestore.iter_next(child_iter)

    return tree

  def on_copy(self, widget=None):
    if self._handle_clipboard(widget, "copy"):
      return

    (model, row) = self.treeview.get_selection().get_selected()
    if row is None:
      return

    tree = etree.ElementTree(self.__get_treestore(row))

    ios = StringIO.StringIO()
    tree.write(ios, pretty_print = True, xml_declaration = False, encoding = "utf-8")

    clipboard = gtk.clipboard_get()
    clipboard.set_text(ios.getvalue())
    clipboard.store()

    ios.close()



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import os
import os.path
import sys
import ConfigParser

import debug

dirs = []
if sys.platform != "win32" and sys.platform != "win64":
  dirs.append("/usr/share/diamond")
  dirs.append("/etc/diamond")
dirs.append(os.path.join(os.path.expanduser('~'), ".diamond"))
if "DIAMOND_CONFIG_PATH" in os.environ:
  dirs += reversed(os.environ["DIAMOND_CONFIG_PATH"].split(":"))

config = ConfigParser.SafeConfigParser()
config.read([os.path.join(path, "settings") for path in reversed(dirs)]) #reversed to load usr last

try:
  config.add_section("colour")
except ConfigParser.DuplicateSectionError:
  pass

def __set_default(option, value):
  if not config.has_option("colour", option):
    config.set("colour", option, value)

__set_default("normal", "black")
__set_default("insert", "green")
__set_default("delete", "red")
__set_default("update", "blue")
__set_default("subupdate", "cornflowerblue")
__set_default("diffadd", "lightgreen")
__set_default("diffsub", "indianred")

schemata = {}

for dir in [os.path.join(path, "schemata") for path in dirs]:
  try:
    for file in os.listdir(dir):
      if file[-1] == "~" or file[0] == ".": #skip files like .nfs0000 
        continue # bloody emacs
      # Skip item gracefully here if there's a problem.
      # This is useful if the schemata files are in a subversion
      # repository and there's pesky .svn folders around.
      try:
        handle = open(os.path.join(dir, file))
      except:
        debug.deprint("Failure to examine entry " + file + " in folder " + dir + ".")
        continue
      lines = [x.strip() for x in handle if x.strip()]
      if len(lines) < 2:
        debug.deprint("Warning: Found schema registration file \"" + file + "\", but file is improperly formatted - schema type not registered", 0)
        continue

      # Expand environment variables in the schema path
      alias = {}
      for i in range(1, len(lines)):
        line = lines[i]

        keyvalue = [x.strip() for x in line.split("=")]
        if len(keyvalue) == 1:
          key, value = ("default", keyvalue[0])
        elif len(keyvalue) == 2:
          (key, value) = keyvalue
        else:
          debug.deprint("Warning: ignoring line %s" % line)
          continue

        value = os.path.expandvars(value)

        if key in alias:
          debug.deprint("""alias "%s" already registered, ignoring""" % key)
        else:
          alias[key] = value
          if key == "default":
            alias[None] = value

      schemata[file] = (lines[0], alias)
      debug.dprint("Registered schema type: " + file)
  except OSError:
    pass

if __name__ == "__main__":
  for key in schemata:
    debug.dprint("%s: %s" % (key, schemata[key]), 0)



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

from cStringIO import StringIO
from lxml import etree

from schema import Schema

RELAXNGNS = "http://relaxng.org/ns/structure/1.0"
RELAXNG = "{" + RELAXNGNS + "}"

def find_fullset(tree):
  """
  Given a schema tree pulls out xpaths for every element.
  """

  cache = set()

  def traverse(node):

    if node.tag == RELAXNG + "element" or (node.tag == RELAXNG + "choice" and all(n.tag != RELAXNG + "value" for n in node)):
      fullset.add(tree.getpath(node))

    elif node.tag == RELAXNG + "ref":
      query = '/t:grammar/t:define[@name="' + node.get("name") + '"]'
      if query in cache:
        return # we only need to add the reference once 
      cache.add(query)
      node = tree.xpath('/t:grammar/t:define[@name="' + node.get("name") + '"]', namespaces={'t': RELAXNGNS})[0]

    for child in node:
      traverse(child)

  start = tree.xpath("/t:grammar/t:start", namespaces={'t': RELAXNGNS})[0]

  root = start[0]

  fullset = set()
  traverse(root)
  return fullset

def find_useset(tree):
  """
  Given a diamond xml tree pulls out scehama paths for every element and attribute.
  """

  def traverse(node):
    if node.active:
      useset.add(node.schemaname)

      for child in node.get_children():
        traverse(child)

  useset = set()
  traverse(tree)
  return useset

def find_unusedset(schema, paths):
  """
  Given the a diamond schema and a list of paths to xml files
  find the unused xpaths.
  """
  def traverse(node):
    if node.active:
      unusedset.discard(node.schemaname)

      for child in node.get_children():
        traverse(child)

  unusedset = find_fullset(schema.tree)

  for path in paths:
    try:
      tree = schema.read(path)
      traverse(tree)
    except IOError:
      pass

  return unusedset

def strip(tag):
  return tag[tag.index("}") + 1:]

def node_name(node):
  """
  Returns a name for this node.
  """
  tagname = node.get("name") if "name" in node.keys() else strip(node.tag)
  name = None

  for child in node:
    if child.tag == RELAXNG + "attribute":
      if "name" in child.keys() and child.get("name") == "name":
        for grandchild in child:
          if grandchild.tag == RELAXNG + "value":
            name = " (" + grandchild.text + ")"
            break
    if name:
      break
  return tagname + (name if name else "")



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import debug
import pango
import gobject
import gtk
import gtk.glade
import sys
import tempfile
import os
import re

import dialogs
import interface

class DiamondSchemaError:
  def __init__(self, parent, gladefile, schema_file, schematron_file):
    self.parent = parent
    self.gladefile = gladefile
    self.schema_file = schema_file
    self.schematron_file = schematron_file
    self.errlist_type = 0
    self.listview = None

    self.listwindow = self.parent.gui.get_widget("errListWindow")
    self.listview = self.parent.gui.get_widget("sch_err_list")

    if self.listwindow is None or self.listview is None:
      raise Exception("Could not find the error list widgets")

    self.model = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
    self.cellcombo = gtk.CellRendererCombo()
    self.cellcombo.set_property("model", self.model)
    self.cellcombo.set_property("editable", False)

    column_renderer = gtk.CellRendererText()
    column_renderer.set_property("editable", False)
    column_xpath = gtk.TreeViewColumn("Element", column_renderer, text = 0)
    column_error = gtk.TreeViewColumn("Error", column_renderer, text = 1)

    self.listview.append_column(column_xpath)
    self.listview.append_column(column_error)

    column_xpath.set_property("expand", True)
    column_xpath.set_resizable(True)
    column_xpath.set_property("min-width", 75)

    column_error.set_property("expand", True)
    column_error.set_resizable(True)
    column_error.set_property("min-width", 75)

    # Set adjustment parameters for the list view
    self.listview.set_model(self.model)
    self.listview.connect("row_activated", self.on_xml_row_activate)

    # User can only select one error at a time to go to.
    self.listview.get_selection().set_mode(gtk.SELECTION_SINGLE)

    self.listview.hide()
    self.listwindow.hide()

  # Create the error list, model and tree view columns if the list does not exist.
  def create_error_list(self):
    if self.listview is not None:
      return

  # Destroy the error list widgets
  def destroy_error_list(self):
    if self.listview is not None:
      self.listview.hide()
      self.listwindow.hide()

  def on_validate(self, widget=None):
    """
    Tools > Validate XML. This writes out the XML to a temporary file, then calls
    xmllint on it. (I didn't use the Xvif validation capabilities because the error
    messages are worse than useless; xmllint actually gives line numbers and such).
    """

    if self.schema_file is None:
      dialogs.error(self.parent.main_window, "No schema file open")
      return

    self.tmp = tempfile.NamedTemporaryFile()
    self.parent.tree.write(self.tmp.name)

    std_input, std_output, std_error = os.popen3("xmllint --relaxng %s %s --noout --path \".\"" % (self.schema_file, self.tmp.name))
    output = std_error.read()

    output = output.replace("%s fails to validate" % self.tmp.name, "")

    if output.strip() == "%s validates" % self.tmp.name:
      # No errors. Close the error list (if it is open) and display a congratulatory message box.
      self.destroy_error_list()
      dialogs.message_box(self.parent.main_window, "XML file validated successfully", "Validation successful")
    else:
      self.create_error_list()

      self.errlist_type = 1
      self.model.clear()

      lines = output.split("\n")

      # Read the temporary output file, split it by lines and use it to convert
      # line numbers to xpaths for the column data.
      f = file(self.tmp.name)
      data = f.read()
      output_lines = data.split("\n")

      for line in lines:
        if len(line) == 0:
          continue

        tokens = line.split(" ")

	# Parse each line of the xmllint --relaxng output.
	# FORMAT:
	# ELEMENT:LINE: element NAME:  Relax-NG validity error : ERROR MESSAGE
	# (Capitals denotes variable data). Update the following code if the output changes.
        sub_tokens = tokens[0].split(":")

        message = " ".join(tokens[7:])

        line = sub_tokens[1]
        xpath = self.add_to_xpath(output_lines, "", long(line)-1)

        self.model.append([ xpath, message ])

      self.listview.set_property("visible", True)
      self.listwindow.set_property("visible", True)

      # Update the status bar. Inform the user of the number of errors.
      self.parent.statusbar.set_statusbar("There are %u Relax-NG errors in the document" % len(lines))

    return

  def on_validate_schematron(self, widget=None):
    """
    Tools > Validate Schematron. This uses the etree.Schematron API, if it exists, to
    validate the document tree against a supplied schematron file.
    """

    if self.schematron_file is None:
      dialogs.error(self.parent.main_window, "No schematron file supplied")
      return

    # Write the current version out to file.
    tmp = tempfile.NamedTemporaryFile()
    self.parent.tree.write(tmp.name)
    std_input, std_output, err_output = os.popen3("xmllint --schematron %s %s --noout" % (self.schematron_file, tmp.name))
    output = err_output.read()

    output = output.replace("%s fails to validate" % tmp.name, "")
    output = output.strip()

    if output == "%s validates" % tmp.name:
      self.destroy_error_list()
      dialogs.message_box(self.parent.main_window, "XML file validated successfully against %s" % self.schematron_file, "XML validation successful")

    # Clear the list, as there may still be errors from a previous schematron validation.
    self.model.clear()
    self.errlist_type = 0

    # Add each line of output as a new row.
    lines = output.split("\n")

    for line in lines:
      if len(line) == 0:
        continue

    tokens = line.split(" ")
    self.model.append([ tokens[0], " ".join(tokens[3:]), 0 ])

    # Finish set up of list view widget.
    self.listview.set_property("visible", True)
    self.listwindow.set_property("visible", True)
    self.parent.statusbar.set_statusbar("There are %u Schematron errors in the document" % len(lines))

  def errlist_is_open(self):
    return self.listview.get_property("visible")

  def get_elem_name(self, line):
    xml = re.compile("<([/?!]?\w+)", re.VERBOSE)
    matches = xml.findall(line)
 #   print "get_elem_name", line, "=", matches[0]
    return matches[0]

  def add_to_xpath(self, lines, xpath, line_no):

    line = lines[line_no].strip()
    name = self.get_elem_name(line)

    if name not in ["string_value", "integer_value", "real_value"]:
      xpath = "/" + name + xpath

    # Are we at the start of the document?
    if line_no == 0 or line_no == 1:
      return xpath

    # Find the parent node.
    while 1:
      line_no = line_no - 1
      line = lines[line_no].strip()

#      print "MAIN LOOP: ", line

      # Skip past comments
      if line.find("-->") != -1:
        while line.find("<!--") == -1:
          line_no = line_no - 1
          line = lines[line_no].strip()
        continue

      if line.startswith("</"):
        name = line.strip("<>/")
        while line.find("<%s" % (name)) == -1:
#          print "BACK LOOP: ", line
          line_no = line_no - 1
          line = lines[line_no].strip()
        continue

      name = self.get_elem_name(line)
      if name.find("<%s />" % (name)) == -1 and name.find("</ %s>" % (name)):
        return self.add_to_xpath(lines, xpath, line_no)

  def scroll_to_xpath(self, xpath):
    iter = self.parent.get_treestore_iter_from_xmlpath(xpath)
    path = self.parent.treestore.get_path(iter)
    self.parent.treeview.expand_to_path(path)
    self.parent.treeview.get_selection().select_iter(iter)
    self.parent.treeview.scroll_to_cell(path, use_align=True, col_align=0.5)

  # Signal handlers
  def on_xml_row_activate(self, treeview, path, view_column):

    # Get the selected row.
    selection = treeview.get_selection()
    (model, iter) = selection.get_selected()

    # Get the line number
    xpath = model.get_value(iter, 0)

    # Use the in-memory schema representation to get the XPath from the name.
    self.scroll_to_xpath(xpath)



from distutils.core import setup
from distutils.extension import Extension
import os
import os.path
import glob

# There are a number of local hacks in this file, to deal with the multiple
# ways in which setup.py is called by various scripts and packaging methods 
# that interact with spud, enabling setuptools to grok their intentions.

# In some cases, we will be passed a 'DESTDIR' from an upstream packagaing
# system. This will be a local directory to install into, and act as local '/'
# as far as all paths are concerned. Check for this, and fail nicely if not set.

prefix = None
import sys
packaging=False

# We may also be given prefix, either as a configuration option (which will be
# dealt with by substitutions later) or as a command line option. If a command
# line option is present, parse this. In some cases, just to add spice into the
# mix, PREFIX is supplied twice - once by configure, and again by a parent
# Makefile as a command-line flag.
#
# As an alternative to DESTDIR being set, Debian packaging calls setup.py with
# a --root command line option which does the same thing. This needs to be
# parsed if present, and supercedes any previous DESTDIR picked up from
# environment.

for i, arg in enumerate(sys.argv):
  if "--prefix" in arg:
    prefix = arg.split('=')[1]

# Given the above prefix possibilities, as well as root and DESTDIR, we need to
# construct a list of data directories to be installed
#
#   * configure-supplied prefixes are dealt with by substitution directly into
#       this file, with @ PREFIX @ being replaced.
#   * prefixes as command-line options are pushed by us
#
# Note that in the case of this and other packages which supply prefix twice
# (once in configure, and then again on the command line) we honour the prefix
# on the command line in preference to the configure prefix.

# First parse the plugin directories
plugin_dirs = []
if os.path.exists('plugins'):
  plugin_dirs = [dir for dir in os.listdir('plugins') if os.path.isdir(os.path.join('plugins', dir)) and dir[0] != '.']
plugin_data_files = []
if sys.platform == 'darwin' and packaging:
	for plugin in plugin_dirs:
  		plugin_data_files.append(("./plugins/" + plugin,
		  glob.glob('plugins/' + plugin + '/*.py')))
else:
	for plugin in plugin_dirs:
	  if prefix is None:
		plugin_data_files.append(("/usr/local/share/diamond/plugins/" + plugin,
		  glob.glob('plugins/' + plugin + '/*.py')))
	  else:
		plugin_data_files.append((prefix + "/share/diamond/plugins/" + plugin,
		  glob.glob('plugins/' + plugin + '/*.py')))

# Now parse the GUI directories
gui_data_files = []
if sys.platform == 'darwin' and packaging	:
  gui_data_files.append(("./gui",
    ["gui/gui.glade", "gui/diamond.svg", "gui/diamond.png"]))
else:
	if prefix is None:
	  gui_data_files.append(("/usr/local/share/diamond/gui",
		["gui/gui.glade", "gui/diamond.svg"]))
	else:
	  gui_data_files.append((prefix + "/share/diamond/gui",
		["gui/gui.glade", "gui/diamond.svg"]))

# We now have all the information we need; run setup.
setup(
      name='spud-diamond',
      version='1.0.3',
      description="Model option preprocessor",
      author = "The ICOM team",
      author_email = "contact@shingleproject.org",
      url = "http://amcg.ese.ic.ac.uk",
      packages = ['diamond'],
      package_dir = {'diamond': 'diamond'},
      scripts=["bin/diamond"],
      data_files = gui_data_files + plugin_data_files
          + [('', ['README.txt', 'COPYING', 'GPL-3'])],
     )




#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

class List:
  def __init__(self, datatype, cardinality=''):
    self.datatype = datatype
    self.cardinality = cardinality

  # The input to call is a string containing a list of elements seperated by "," or " ". It
  # returns a string, containing the elements in val seperated by " ".
  def __call__(self, val):
    val = val.strip()
    if "," in val:
      x = val.split(",")
    else:
      x = val.split(" ")

    # Perform some checks on the list cardinality, for oneOrMore elements ('+')
    # and compulsory elements ('').
    if self.cardinality == '+':
      assert len(x) > 0
    elif self.cardinality == '':
      assert len(x) == 1

    # Check the list cardinality (as an integer) matches up with the number of elements
    # in the seperated val string.
    try:
      assert len(x) == int(self.cardinality)
    except ValueError:
      pass # The int conversion may fail (if cardinality is '+' or ''), so just ignore it.

    # Make sure each element can be converted to type 'self.datatype'. An exception will be
    # thrown if this is not possible.
    for y in x:
      z = self.datatype(y)

    # Return a string of the elements in val, seperated by " ".
    return " ".join(x)

  def __str__(self):
    return "list of " + str(self.datatype) + " of cardinality: " + self.cardinality

  def __repr__(self):
    return "list of " + str(self.datatype) + " of cardinality: " + self.cardinality



#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import traceback

import gtk

import pygtkconsole

def prompt(parent, message, type = gtk.MESSAGE_QUESTION, has_cancel = False):
  """
  Display a simple Yes / No dialog. Returns one of gtk.RESPONSE_{YES,NO,CANCEL}.
  """

  prompt_dialog = gtk.MessageDialog(parent, 0, type, gtk.BUTTONS_NONE, message)
  prompt_dialog.add_buttons(gtk.STOCK_YES, gtk.RESPONSE_YES, gtk.STOCK_NO, gtk.RESPONSE_NO)
  if has_cancel:
    prompt_dialog.add_buttons(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
  prompt_dialog.connect("response", prompt_response)

  prompt_dialog.run()

  return prompt_response.response

def long_message(parent, message):
  """
  Display a message prompt, with the message contained within a scrolled window.
  """

  message_dialog = gtk.Dialog(parent = parent, buttons = (gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))
  message_dialog.set_default_size(400, 300)
  message_dialog.connect("response", close_dialog)

  scrolled_window = gtk.ScrolledWindow()
  message_dialog.vbox.add(scrolled_window)
  scrolled_window.show()

  scrolled_window.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)

  text_view = gtk.TextView()
  scrolled_window.add(text_view)
  text_view.show()

  text_view.get_buffer().set_text(message)
  text_view.set_cursor_visible(False)
  text_view.set_property("editable", False)
  text_view.set_property("height-request", 180)
  text_view.set_property("width-request", 240)

  message_dialog.run()

  return

def error(parent, message):
  """
  Display an error message.
  """

  error_dialog = gtk.MessageDialog(parent, 0, gtk.MESSAGE_WARNING, gtk.BUTTONS_OK, message)
  error_dialog.connect("response", close_dialog)
  error_dialog.run()

  return

def error_tb(parent, message):
  """
  Display an error message, together with the last traceback.
  """

  tb = traceback.format_exception(sys.exc_info()[0] ,sys.exc_info()[1], sys.exc_info()[2])
  tb_msg = ""
  for tbline in tb:
    tb_msg += tbline
  long_message(parent, tb_msg + "\n" + message)

  return

def get_filename(title, action, filter_names_and_patterns = {}, folder_uri = None):
  """
  Utility function to get a filename.
  """

  if action == gtk.FILE_CHOOSER_ACTION_SAVE:
    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_SAVE,gtk.RESPONSE_OK)
  elif action == gtk.FILE_CHOOSER_ACTION_CREATE_FOLDER:
    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_NEW,gtk.RESPONSE_OK)
  else:
    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK)
  filew = gtk.FileChooserDialog(title=title, action=action, buttons=buttons)
  filew.set_default_response(gtk.RESPONSE_OK)

  if not folder_uri is None:
    filew.set_current_folder_uri("file://" + os.path.abspath(folder_uri))

  for filtername in filter_names_and_patterns:
    filter = gtk.FileFilter()
    filter.set_name(filtername)
    filter.add_pattern(filter_names_and_patterns[filtername])
    filew.add_filter(filter)

  allfilter = gtk.FileFilter()
  allfilter.set_name("All known files")
  for filtername in filter_names_and_patterns:
    allfilter.add_pattern(filter_names_and_patterns[filtername])
  filew.add_filter(allfilter)

  filter = gtk.FileFilter()
  filter.set_name("All files")
  filter.add_pattern("*")
  filew.add_filter(filter)
 
  result = filew.run()

  if result == gtk.RESPONSE_OK:
    filename = filew.get_filename()
    filtername = filew.get_filter().get_name()
    filew.destroy()
    return filename
  else:
    filew.destroy()
    return None
    
def console(parent, locals = None):
  """
  Launch a python console.
  """

  console_dialog = gtk.Dialog(parent = parent, buttons = (gtk.STOCK_QUIT, gtk.RESPONSE_ACCEPT))
  console_dialog.set_default_size(400, 300)
  console_dialog.connect("response", close_dialog)

  stdout = sys.stdout
  stderr = sys.stderr

  console_widget = pygtkconsole.GTKInterpreterConsole(locals)
  console_dialog.vbox.add(console_widget)
  console_widget.show()

  console_dialog.run()
  
  sys.stdout = stdout
  sys.stderr = stderr
  
  return

def prompt_response(dialog, response_id):
  """
  Signal handler for dialog response signals. Stores the dialog response in the
  function namespace, to allow response return in other functions.
  """

  if response_id == gtk.RESPONSE_DELETE_EVENT:
    response_id = gtk.RESPONSE_CANCEL

  prompt_response.response = response_id
  close_dialog(dialog, response_id)

  return

def close_dialog(dialog, response_id = None):
  """
  Signal handler for dialog reponse or destroy signals. Closes the dialog.
  """

  dialog.destroy()

  return

def radio_dialog(title, message, choices, logo):
  r = RadioDialog(title, message, choices, logo)
  return r.data

def message_box(window, title, message):
  dialog = gtk.MessageDialog(window, 0, gtk.MESSAGE_INFO, gtk.BUTTONS_OK, message)
  dialog.set_title(title)
  dialog.connect("response", close_dialog)
  dialog.run()

class RadioDialog:
  def __init__(self, title, message, choices, logo):
    self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    self.window.connect("delete_event", self.cleanup)
    self.window.connect("key_press_event", self.key_press)
    self.window.set_title(title)
    self.window.set_position(gtk.WIN_POS_CENTER)
    if not logo is None:
      self.window.set_icon_from_file(logo)
    self.window.show()

    #swindow = gtk.ScrolledWindow()
    #swindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    #self.window.add(swindow)
    #swindow.show()

    main_box = gtk.VBox(False, 0)
    self.window.add(main_box)
    main_box.show()

    if not logo is None:
      image = gtk.Image()
      image.set_from_file(logo)
      main_box.pack_start(image, True, True, 0)
      image.show()

    label = gtk.Label(message)
    main_box.add(label)
    label.show()

    radio_box = gtk.VBox(False, 10)
    main_box.pack_start(radio_box, True, True, 0)
    radio_box.show()

    separator = gtk.HSeparator()
    main_box.pack_start(separator, False, True, 0)
    separator.show()

    close = gtk.Button(stock=gtk.STOCK_OK)
    close.connect("clicked", self.cleanup)
    main_box.pack_start(close, False, False, 0)
    close.show()

    prev_radio = None
    for choice in choices:
      radio = gtk.RadioButton(prev_radio, choice)
      radio.connect("toggled", self.radio_callback, choice)
      radio_box.pack_start(radio, False, False, 0)
      radio.show()
      if prev_radio is None:
        radio.set_active(True)
      prev_radio = radio

    self.data = choices[0]

    gtk.main()

  def cleanup(self, widget, data=None):
    self.window.destroy()
    gtk.main_quit()

  def key_press(self, widget, event):
    if event.keyval == gtk.keysyms.Return:
      self.cleanup(None)

  def radio_callback(self, widget, data):
    self.data = data

class GoToDialog:
  def __init__(self, parent):
    self.goto_gui = gtk.glade.XML(parent.gladefile, root="GoToDialog")
    self.dialog_box = self.goto_gui.get_widget("GoToDialog")
    self.dialog_box.set_modal(True)

  def run(self):
    signals =      {"goto_activate": self.on_goto_activate,
                    "cancel_activate": self.on_cancel_activate}

    self.goto_gui.signal_autoconnect(signals)

    self.dialog_box.show()
    return ""

  def on_goto_activate(self, widget=None):
    print "goto"

  def on_cancel_activate(self, widget=None):
    print "cancel"




#!/usr/bin/env python

#    This file is part of Diamond.
#
#    Diamond is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Diamond is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Diamond.  If not, see <http://www.gnu.org/licenses/>.

"""
Debugging module. Stores the program debug level and provides debugging output
functions. Only debugging output with level less than or equal to the current
debug level is output. Note that debugging output with level greater than the
current maximum debug level is treated as having a level equal to the current
maximum debug level, and that debugging output with level less than zero is
treated as having a level equal to zero.
"""

import sys

class DebugLevel:
  """
  Class used to store a debug level.
  """

  def __init__(self, level = 1, maxLevel = 3):
    """
    Initialise a debug level.
    """

    self.SetLevel(level)
    self.SetMaxLevel(maxLevel)

    return

  def GetLevel(self):
    """
    Get the debug level.
    """

    return self._level

  def SetLevel(self, level = 1):
    """
    Set the debug level.
    """

    level = max(level, 0)

    try:
      self._level = min(level, self.GetMaxLevel())
    except:
      self._level = level

    return

  def GetMaxLevel(self):
    """
    Get the maximum debug level.
    """

    return self._maxLevel

  def SetMaxLevel(self, maxLevel = 3):
    """
    Set the maximum debug level.
    """

    maxLevel = max(maxLevel, 0)

    self._maxLevel = maxLevel
    self.SetLevel(self._level)

    return

# Stores the current module debug level
_debugLevel = DebugLevel()

def GetDebugLevel():
  """
  Get the current debug level.
  """

  return _debugLevel.GetLevel()

def SetDebugLevel(level = 1):
  """
  Set the current debug level.
  """

  _debugLevel.SetLevel(level)

  return

def GetMaxDebugLevel():
  """
  Get the current maximum debug level.
  """

  return _debugLevel.GetMaxLevel()

def SetMaxDebugLevel(level = 3):
  """
  Set the current maximum debug level.
  """

  _debugLevel.SetMaxLevel(level)

  return

def dprint(msg, level = 1, newline = True, flush = True):
  """
  Print a debug message to standard output with supplied debug level.
  """

  dwrite(sys.stdout, msg, level, newline, flush)

  return

def deprint(msg, level = 1, newline = True, flush = True):
  """
  Print a debug message to standard error with supplied debug level.
  """

  dwrite(sys.stderr, msg, level, newline, flush)

  return

def dwrite(stream, msg, level = 1, newline = True, flush = True):
  """
  Print a debug message to the supplied file stream with supplied debug level.
  """

  level = max(level, 0)
  level = min(level, GetMaxDebugLevel())

  if level <= GetDebugLevel():
    stream.write(str(msg))
    if newline:
      stream.write("\n")
    if flush:
      stream.flush()

  return
